import regex
from framaInterface import getFunctionNamesUsedByFunction, getEdgesRecursive
import re
import logging



class Function(object):
    """a data class, to contain the function definition data"""
    def __init__(self,static:str, returnType:str, functionName: str, arguments: str, implementationCode: str, entireCode: str, filePath: str):
        self.static = static #if the function in static
        self.returnType = returnType # the return type of the function
        self.functionName = functionName # the name of the function
        self.arguments = arguments #the arguments of the function
        self.implementationCode = implementationCode #the code inside the { ... } 
        self.entireCode = entireCode #the all code of the function 
        self.filePath = filePath # the path of the file where the function is define. 
        self.__calledFunctions__ = None # set only as needed
        self.__calledEdges__ = None # set only as needed

        self.sgx_edl_argument_definition = None # set only if ecall/ocall function

    def getSgxForEdl(self):
        """return the argument definition for edl file"""
        return  self.sgx_edl_argument_definition

    def setSgxArgumentDefinition(self, sgx_argument_macro_definition: str ): 
        if self.static:
            raise Exception(f"Error #F001: ecall and ocall can't be defined static. Static is found for function: {self.static} {self.functionName}{self.arguments}")
        logging.debug(f"inspection of the macro function of: {self.functionName}")
        
        if self.returnType != "int": #check if ecall has return type = int
            raise Exception(f"Error #F002: return type of ecall {self.functionName} is not int, found: {self.returnType}""")

        self.__check_compute_sgx_arguments__(sgx_argument_macro_definition)

    def __check_compute_sgx_arguments__(self, sgx_argument_macro_definition):
        """expected format like ([name, type,  size],..)\n
        *name* is the name of a argument\n
        *type* can be i,o,b,u. one single character only. i = in, o = out, b= in,out, u=user_check\n
        *size* can be void or, a number or the name of an integer variable\n"""

        argumentsToSpecify = 0        

        realArgumentDict = {}
        for realArg in self.getArguments():
            realArgumentDict[realArg['argName']] = realArg

            if realArg['dynamicArray'] > 0:
                raise Exception ("Error #F003: Ecall and Ocall can't use dynamic array sintax, use pointer * instead. [] -> *")


            if realArg["const"] and realArg['pointers'] == 0:
                raise Exception ("Error #F004: Ecall and Ocall can't have const arguments that are not pointers")

            if (realArg['staticArray'] + realArg['pointers']) > 0:
                argumentsToSpecify += 1

        sgxArgList = re.findall(r"\[([^\[\]]*)\]", sgx_argument_macro_definition)

        if len(sgxArgList) != argumentsToSpecify:
            raise Exception(f"Error #F005: wrong number of specified elemens in sgx_macro. Excpected {argumentsToSpecify} but is {len(sgxArgList)}: for function {self.functionName}")


        for arg in sgxArgList:
            match = re.match(r"[ ]*([\w\d]+)[ ,]+([iobu])(?:[ ,]+([\w\d]+))?[\s]*$", arg)

            if not match:
                raise Exception(f"""Error #F006: the sgx argument definition of function: {self.functionName} has errors in def: {[arg]}""")
                

            #match
            argName, copyOption, size = match.groups()    

            #check size definition 
            if size: # has size definition
                if not size.isnumeric(): # check that the size name exist in args
                    if size not in realArgumentDict:
                        raise Exception(f"Error #F007: the sgx size definition {size} is not found in function {self.functionName}")
                    
                    sizeArg = realArgumentDict[size] 

                    if sizeArg['typeName'] != "int":
                        raise Exception(f"Error #F008: the size argument of sgx macro must be of tipe int. in macro  {arg} for function {self.functionName}")
                    elif sizeArg['pointers'] != 0 and sizeArg['dynamicArray'] != 0 and sizeArg['staticArray'] != 0:
                        raise Exception(f"Error #F009: the size argument of sgx macro must be of type int without pointers or array type. in macro  {arg} for function {self.functionName}")
            

            # checking that argument exist
            realArg = None
            realArg = realArgumentDict.get(argName)
            if not realArg:
                raise Exception(f"Error #F010: the sgx argument defintion {argName} is not found in function {self.functionName}")


            # check that the argument if i/o/b definition
            if copyOption in ['i','o','b']: #in out in/out selected
                pointersLevel =  realArg["pointers"] + realArg["dynamicArray"]
                
                if pointersLevel == 0 and size is not None: # no sense to specify the size of a non pointer object 
                    raise Exception(f"Error #F011: the size option must be specified only for pointers: in {self.functionName} in variable: {argName}")
                    
                if pointersLevel == 1 and size is None: # the size must be specified with pointes
                    raise Exception(f"Error #F012: the size option must be specified with pointers: in {self.functionName} in variable: {argName}")

                if pointersLevel > 1: # more than one pointer level = no possibility to copy correctly in and out 
                    raise Exception(f"Error #F013: multiple pointer type like (var** or  *var[]) can used only with 'u' copy option in sxg macro definition of function: {self.functionName} in variable: {argName}  ")

                #todo check the struct and unions has no pointer inside

            elif copyOption == 'u': # check the user_check option
                if size is not None: # the size must be none with user_check option
                    raise Exception(f"Error #F014: the size must be none is user_check option is selected. in: {self.functionName} in variable: {argName}  ")
            else:
                raise Exception(f"Error #F015: option {copyOption} not available for macro function {self.functionName}")

            copyOption = {"i":"in", "o":"out", "b":"in, out", "u":"user_check"}[copyOption] #transform the single char in the correct definition
            
            #computing size notation
            if size is None:
                size = ""
            elif realArg['typeName'] == "void": # use size
                size = f", size={size}"
            else: # use count notation for defined obgect instead 
                size = f", count={size}" 
                
            realArg["completeDef"] = f"""[{copyOption}{size}] {realArg["completeDef"]} """    # compute edl syntax
       
        
        eldArgumentList = list(arg['completeDef'] for arg in realArgumentDict.values()) #return the list of arguments with edl syntax
        self.sgx_edl_argument_definition  = "(" + ", ".join(eldArgumentList) + ")" # compute edl syntax
            

    def getPrototype(self)-> str:
        return f"""{self.returnType} {self.functionName}{self.arguments};\n"""

    def __str__(self):
        return self.entireCode

    def getCalledFunctioNames(self, recursive)-> []:
        """return a list of functions names reachable by the root self function, if recursive = False, recursion level = 1"""
        if self.__calledFunctions__ == None:
            self.__calledFunctions__ = getFunctionNamesUsedByFunction(self.functionName, self.filePath, recursive)
        
        return self.__calledFunctions__
    
    def getFunctionCalledEdgesRecursive(self) -> []:
        """get a list of [{source: functName1, dest: funcName2}] of function reached by the self function"""
        if self.__calledEdges__ == None:
            self.__calledEdges__ = getEdgesRecursive(self.functionName, self.filePath)

        return self.__calledEdges__

    def getArguments(self)-> []: 
        """maybe better perfomance, once computed never computed again #TODO"""
        """ return a list of argument of the function """
        args = self.arguments
        args = args.strip("(").strip(")").split(",")

        ret = []
        for arg in args:
            arg = arg.strip(" ")
            isStructEnumUnion = re.match(r"^[ ]*(struct|enum|union)[\s]+([a-zA-Z_]*)([\s*]+)([a-zA-Z_]*)", arg)
            isIntVoidChar  = re.match(r"^[ ]*(const[\s]+)?(unsigned[\s]+)?(int|void|short|char|float|double)([\s*]+)([a-zA-Z_]+)", arg)

            if isStructEnumUnion:
                argType, typeName, pointers, argName = isStructEnumUnion.groups()
            
            elif isIntVoidChar:
                argType = "simple"
                const, unsigned, typeName, pointers, argName = isIntVoidChar.groups()
                const = const.strip(" ") if const else None
                unsigned = unsigned.strip(" ") if unsigned else None
            elif "void" in arg or not arg:
                break # ok just void argument
            else:
                raise Exception(f"""Error #F016: {self.functionName}({arg}) type not supported by framework)""")
            
            pointers = pointers.count("*")

            dynamicArray = len(re.findall(r"(\[[\s]*\])", arg))
            staticArray = len(re.findall(r"(\[[\s]*(?:[\d]+|[\w]+)[\s]*\])", arg))

            ret.append({"completeDef": arg, 
            "argType": argType, 
            "typeName": typeName, 
            "pointers": pointers, 
            "argName": argName, 
            "dynamicArray": dynamicArray, 
            "staticArray": staticArray,
            "const": const,
            "unsigned": unsigned})
            
        return ret


    @staticmethod
    def getNamesOfFunctionList(functions: []):
        functionNames = []
        for f in functions:
            functionNames.append(f.functionName)
        return functionNames

    @staticmethod
    def splitFunctionToEcallWrapperAndEcallTrusted(ecall):
        """ modify the ecall ecallTrusted and generate the ecallWrapper create new name for the ecallTrusted
        create the implementation of EcallWrapper that calls the EcallTrusted and manages the errors. 
        returns the ecallWrapper, the trusted is modified from argument"""
        
        ecallTrustedNewName = ecall.functionName + "_trusted"

# ------------- ecall wrapper creation -----------------
        ecallWrapperFunctionc: Function = Function("", ecall.returnType, ecall.functionName, ecall.arguments, "implementation Todo", "entireCode Todo", ecall.filePath)

        argumentsNames: [] = ", ".join([a['argName'] for a in ecall.getArguments()]) # compute the list of arguments divided by comma
        ecallWrapperFunctionc.implementationCode = f"""{{   
    create_enclave();
    {ecall.returnType} retval;
    sgx_status_t ret = {ecallTrustedNewName}(my_enclave_id, &retval{", " if len(argumentsNames) else "" }{argumentsNames}); 
    if (ret != SGX_SUCCESS) {{
        printf("Enclave error");
        exit(-1);
    }}     
    return retval;
}}"""
        ecallWrapperFunctionc.entireCode = f"""{ecallWrapperFunctionc.returnType} {ecallWrapperFunctionc.functionName} {ecallWrapperFunctionc.arguments} {ecallWrapperFunctionc.implementationCode} """  

# -------------- ecall trusted creation -----------------
        ecallTrusteFunction = Function(ecall.static, ecall.returnType, ecallTrustedNewName, ecall.arguments, ecall.implementationCode, ecall.entireCode.replace(ecall.functionName, ecallTrustedNewName, 1), "" )
        ecallTrusteFunction.sgx_edl_argument_definition = ecall.sgx_edl_argument_definition
        return ecallTrusteFunction, ecallWrapperFunctionc

    @staticmethod
    def generateOcallWrapper(ocall, filePath: str):
        """create the implementation to call the ocall and check the return status, filePath is path of the file where the function will be placed in """
        ocallWrapper: Function = Function("", ocall.returnType, ocall.functionName+"_wrapper", ocall.arguments, "impleCodeTODO", "entireCodeTODO", filePath)

        argumentsNames: [] = ", ".join([a['argName'] for a in ocall.getArguments()]) # compute the list of arguments divided by comma

        
        ocallWrapper.implementationCode = f"""{{
    int retval;
    sgx_status_t ret = {ocall.functionName}(&retval{", " if len(argumentsNames) else "" }{argumentsNames});
    if(ret != SGX_SUCCESS){{
        //printf("Enclave error in ocall function: {ocall.functionName}");
        abort();
    }}
    return retval;
}}
#define {ocall.functionName} {ocallWrapper.functionName}
"""

        ocallWrapper.entireCode = f"""{ocallWrapper.returnType} {ocallWrapper.functionName}{ocallWrapper.arguments} {ocallWrapper.implementationCode}"""
        return ocallWrapper


    
#try yourself
#f: Function = Function("int", "foo", "(  char** * buff , int len, struct ciao**la, enum piso rin )", "{return;}", "", "")
#for arg in f.getArguments():
#    print(f"""
# argType  = {arg["argType"]}
# typeName = {arg["typeName"]}
# pointers = {arg["pointers"]}
# argName  = {arg["argName"]}""")
