from ctagsInterface import c_types
import ctagsInterface
from cFileReaderWriter import cFileReaderWriter
from Function import Function
import regex
from constants import Constants
import gccInterface
from tqdm import tqdm #https://github.com/tqdm/tqdm#installation
from EdlFileWriter import EdlFileWriter
import framaInterface

import logging, sys
import os, shutil
import time

start_time = time.time()

log_format = "%(levelname)-8s: %(filename)-20s.%(lineno)-3d - %(message)s"
#logging.basicConfig(stream=sys.stderr, level=logging.DEBUG, format=log_format) #CRITICAL 50, ERROR 40, WARNING 30, INFO 20, DEBUG 10, NOTSET 0
logging.basicConfig(filename="logfile.txt", filemode='w', level=logging.DEBUG, format=log_format) #CRITICAL 50, ERROR 40, WARNING 30, INFO 20, DEBUG 10, NOTSET 0
        

def main(filesPaths: [], src, ):

    checkToolsAvailable()

# ------ check part start -------------------------------------------------------------------------

    dst = "generated_trusted"
    copyRemoveAndCopyTree(src, dst)
    logging.info(f"\n\nCoverting code from {src} to {os.getcwd()}")
    logging.info("Working in directory: " + os.getcwd())

    checkFilesExistance(filesPaths)
    checkAllFileCOrH(filesPaths)

    logging.info("checking code portability")

    ctagsFunction = ctagsInterface.getNamesOfType(filesPaths, c_types.function) # assert all functions have unique names
    structsDef = ctagsInterface.getNamesOfType(filesPaths, c_types.struct) #same
    enumsDef = ctagsInterface.getNamesOfType(filesPaths, c_types.enum) #same
    unionsDef = ctagsInterface.getNamesOfType(filesPaths, c_types.union) #same
    definesDef = ctagsInterface.getNamesOfType(filesPaths, c_types.define) #same


    logging.debug("reading and parsing all files" + str(filesPaths))

    cFileReaderWriterDict = {}
    for filePath in filesPaths: # open and parse all files
        fileReaderWriter = cFileReaderWriter(filePath)  # parse the file
        cFileReaderWriterDict[filePath] = fileReaderWriter

    
    # get ecallFunctions ocallFunctions allFunctions 
    ecallFunctions, ocallFunctions, allFunctions = getEcallsOcallsAndAllFunc(cFileReaderWriterDict.values())
    logging.info("found ecall: " + str(Function.getNamesOfFunctionList(ecallFunctions.values())))
    logging.info("found ocall: " + str(Function.getNamesOfFunctionList(ocallFunctions.values())))

    if len(ecallFunctions)==0:
        raise Exception("Error #B001: there are no ecall definition")

    if len(ctagsFunction) != len(allFunctions): #check we read all files correctly
        raise Exception("Error #B002: something is not working with ctags and framework regex. Are there any non C11 code?")

    assertDifferentNamesOcallEcall(ecallFunctions, ocallFunctions) # check if an ecall and ocall has different names
    
    checkArgumentsTypeDefinitionInFunctions(structsDef, enumsDef, unionsDef,  ecallFunctions) #check if the definition of user type exists
    checkArgumentsTypeDefinitionInFunctions(structsDef, enumsDef, unionsDef,  ocallFunctions) #check if the definition of user type exists
    

    librariesToInclude, enclaveFunctions = findAllTrustedCodeAndLibrary(ecallFunctions, ocallFunctions, allFunctions, filesPaths) # enclave functions are function to move into enclave
    checkSeparationOfContext(ecallFunctions, enclaveFunctions, ocallFunctions, allFunctions)


    #find the ecall needed function
    ecallNeededFunctions: dict = [f for f in enclaveFunctions if f not in ecallFunctions] 
    
    enclaveFunctions = None

# ------ creating / modifing part ----------------------------------------------------------------------------------

    os.mkdir("../enclave")


    # creating enclave.c file
    logging.debug("generation of enclave.c")
    enclaveC = cFileReaderWriter("../enclave/Enclave.c", new = True)


    logging.debug("starting modifing the files")

    #make ocall wrappers
    for ocallName in ocallFunctions:
        ocall = allFunctions[ocallName]
        OcallWrapper: Function = Function.generateOcallWrapper(ocall, enclaveC.filePath)

        enclaveC.addFunction(OcallWrapper)

    #adding ecall used function to the enclave.c
    for funName in ecallNeededFunctions:
        fun = allFunctions[funName]
        cFileReaderWriterDict[fun.filePath].popFunction(fun.functionName) # pop the function from the original file
        enclaveC.addFunction(fun) # add all ecall needed function in the file 
    
    # split the ecalls in the caller(untrusted wrapper) and callee(trusted) and update the files
    for ecallName in ecallFunctions:
        ecall = allFunctions[ecallName]
        fileWhereEcallIsDef: cFileReaderWriter = cFileReaderWriterDict[ecall.filePath]  # get the file where function is def
        
        ecallTrustedFunc, ecallWrapperFunc = Function.splitFunctionToEcallWrapperAndEcallTrusted(ecall) # split the function in wrapper and ecallTrusted       
        fileWhereEcallIsDef.replaceFunction(ecall, ecallWrapperFunc) #replace with the new one

        enclaveC.addCodeToTheStart(f"""#define {ecallName} {ecallTrustedFunc.functionName}""")
        
        enclaveC.addFunction(ecallTrustedFunc) # add the trusted ecall function in the enclave.c file
    ecallFunctions = None # Safety, this list has no sense anymore since we split the functions

    #copy / move code requested by user
    enclaveC.addCodeToTheStart(getCopyMoveCodes(cFileReaderWriterDict))

    #adding #define to enclave.c
    enclaveC.addCodeToTheStart(getAllDefineCode(cFileReaderWriterDict))     

    #adding include to the enclave.c
    logging.debug("adding include to enclave.c")
    enclaveC.addInclude("Enclave_t.h")
    enclaveC.addInclude("stdlib.h")
    for library in librariesToInclude:
        enclaveC.addInclude(library)


    enclaveC.saveFile()

# ---------------- edl file creation --------------------- 
    logging.debug("pop all the user type code")
    typeDefinitionsCode = popAllTypedefinitionsAndGetTheCode(cFileReaderWriterDict, "enclave_manager.h") # get all the struct, enum, union definition code
    
    logging.debug("creation of enclave.edl")
    edlFile = EdlFileWriter("../enclave/Enclave.edl") # edl file writer object
    
    logging.debug("pushing popped code to the edl file")
    edlFile.setTypeDefinitions(typeDefinitionsCode) # add the type definition to the edl

    #define ecall definitions in edl file
    for enclaveFunctions in enclaveC.getAllFunctions().values():
        if enclaveFunctions.sgx_edl_argument_definition: #has macro definition
            edlFile.addTrustedFunction(enclaveFunctions)

    #define ocall definitions in edl file
    for ocallName in ocallFunctions:
        ocall = allFunctions[ocallName]
        edlFile.addUntrustedFunction(ocall)

    #adding #define to Enclave.edl
    edlFile.addDefines(getAllDefineCode(cFileReaderWriterDict))     

    edlFile.writeEdlFile()

# ---------- creating enclave_manager.h -----------------
    Constants.writeEnclaveManager("enclave_manager.h")
    
#----------- save all files -----------------------------
    for filePath, file in cFileReaderWriterDict.items():
        file.saveFile()

# ---------- last operations on logfile -----------------
    logging.info(f"Total elapsed time {round(time.time() - start_time,2)} seconds")
    logging.shutdown()
    shutil.move("../../logfile.txt", "..")

    
def getCopyMoveCodes(cFileReaderWriterDict: dict)-> str: 
    """return a string rappresenting the code to move or copy in the enclave"""
    
    code = ""
    for cFile in cFileReaderWriterDict.values():
        code += cFile.getSGXCopyMoveCode()

    return code


def popAllTypedefinitionsAndGetTheCode(cFileReaderWriterDict: dict, toInclude):
    """return a list of strings rappresenting the code popped from the files.
    
    Pop the definition of the struct, enum, unions from the file and include the new header
    """
    typeDefinitionsCode = []
    for cFile in cFileReaderWriterDict.values():
        typeDefinitionsCode += cFile.popAllDataStructureDefinition()
        cFile.addInclude(toInclude)

    return typeDefinitionsCode

def getAllDefineCode(cFileReaderWriterDict: dict):
    """return a string rappresenting the code of all defines in files. (#define ...)"""
    
    defineCode = []
    for cFile in cFileReaderWriterDict.values():
        defineCode += cFile.getDefines()

    allDefinesString = ""

    for defineStr in defineCode:
        allDefinesString += defineStr + "\n"
    
    return allDefinesString
    


def getEcallsOcallsAndAllFunc(cFileReaderWritersList: []):
    """open and parse all files, return 3 list of functions: ecalls, ocalls, allFuncs.
    Remove the macro annotation """

    ecallFunctions = {} # dict of object Function
    ocallFunctions = {} # dict of object Function
    allFunctions = {} # dict of object Function

    for fileReaderWriter in cFileReaderWritersList:        
        ecalls, ocalls = fileReaderWriter.getEcallOcallFunctions()

        ecallFunctions.update(ecalls)
        ocallFunctions.update(ocalls)
        ecalls, ocalls = None, None # error pruning

        allFunctions.update(fileReaderWriter.getAllFunctions())
    
    return ecallFunctions, ocallFunctions, allFunctions


def findAllTrustedCodeAndLibrary(ecallFunctions: dict, ocallFunctions: dict, allFunctions: dict, filesPaths: [] ):
    """return librariesToInclude: list of dict, functionsToMoveInEnclave: dict of functions with names, also ecall"""
    
    logging.debug("searching ecall called functions")
    functionsNamesToMoveInEnclave = set() # the set of functions names that need to be moved in enclave, that are not ecall

    #find all functions that need to be trusted code
    librariesToInclude = set() # the libraris that must be included in enclave
    for ecallName, ecall in ecallFunctions.items(): #for each ecall
        ecallReachableEdges = ecall.getFunctionCalledEdgesRecursive() # get the edges reachable: ret [{source: functName1, dest: funcName2}]

        #init of algorithm
        seen = set()
        todo = set()
        todo.add(ecallName)
        
        #algorith of graph exploring:  breadth-first search, BFS
        while todo:
            now = todo.pop()
            seen.add(now)
            todo.update(edge["dest"] for edge in ecallReachableEdges  if edge["source"] == now)
            todo -= seen
            todo -= ocallFunctions.keys() #explore only the function not used by ocalls
        
        functionsNamesToMoveInEnclave.update(seen) #the list "seen" rappresent the functions used by the ecall
        logging.debug(f"""the functions {seen} are used by ecall {ecallName}, they will be moved to trusted code""")

        # find the libraries where these functions are defined
        libraryFunctions = seen - allFunctions.keys() #these are the functions used by this ecall, not defined by the user,
        libraries = set(gccInterface.getHeadersWhereFunctionsAreDef(libraryFunctions, filesPaths).values()) # this are the libraries where these functions can be found
        librariesToInclude.update(libraries)
        
        logging.debug(f"""the standard library functions {libraryFunctions} are used by ecall {ecallName}, {libraries} are included""")

        # the set contains only the set of functions user def used by ecalls that are not not ecalls
        functionsNamesToMoveInEnclave -= libraryFunctions # keep only user def functions in the set
    
    functionsToMoveInEnclave = {}
    for functionNameToMove in functionsNamesToMoveInEnclave:
        functionsToMoveInEnclave[functionNameToMove] = allFunctions[functionNameToMove]

    logging.info(f"""the libraries used by ecalls are {librariesToInclude}, they will be included""")

    return librariesToInclude, functionsToMoveInEnclave

def checkTheseNeverCallThose(these: {}, those: {}):
    print("\n\ncheck result")
    for this in tqdm(these.values(), leave=True, bar_format='{l_bar}{bar}| time: {elapsed}'): # this is for the loading bar
        usedBythis = this.getCalledFunctioNames(recursive=False) # get functions names used directly this code
        for that in usedBythis:
            if that in those:
                raise Exception(f"""Error #B003: the untrusted code and trusted code use a same function,
            The function {this} can't use the function {that}. """)

def checkSeparationOfContext(ecallsFunctions, enclaveFunctions, ocallsFunctions, allFunctions):
    """check that the enclave functions never call untrusted functions unless ocalls \n
    check that the untrusted functions never calls enclave functions unless ecalls"""

    trustedFunctions = {}   # all trusted functions that are not ecalls
    untrustedFunctions = {} # all untrusted functions that are not ocalls
       
    for functionName, function in allFunctions.items():
        if functionName in enclaveFunctions and functionName not in ecallsFunctions: # is trusted but no ecall
            trustedFunctions[functionName] = function
        elif functionName not in enclaveFunctions and functionName not in ocallsFunctions: # is untrusted but not ocall 
            untrustedFunctions[functionName] = function

    #check untrusted functions never call trusted functions unless ecalls
    logging.debug("check trusted function never calls untrusted ones")
    checkTheseNeverCallThose(these={**untrustedFunctions, **ocallsFunctions}, those=trustedFunctions)

    #check enclave functions never call untrusted functions unless ocalls
    logging.debug("check untrusted function never calls trusted ones")
    checkTheseNeverCallThose(these={**enclaveFunctions, **ecallsFunctions}, those=untrustedFunctions)

    logging.info(f"""ok, the functions inside the enclave are {list(trustedFunctions.keys()) + list(ecallsFunctions.keys())}, they will be moved""")
    logging.info(f"""ok, the functions outside the enclave are {list(untrustedFunctions.keys()) + list(ocallsFunctions.keys())}, they are kept""")

def checkArgumentsTypeDefinitionInFunctions(structs, enums, unions, functions: {}):
    """check ecall arguments type values.
    functions argument is an array of object Function. 
    check if the definitions of struct, enum, unions exist"""

    for function in functions.values(): #for each function
        logging.debug(f"""searching arguments definition of {function.functionName}""")
        for arg in function.getArguments():             #for each argument
            argType = arg["argType"]
            typeName = arg["typeName"]
            pointers = arg["pointers"]
            argName = arg["argName"] 

            if argType != "simple": #check if the user def type is well defined
                found = False
                if argType == "struct":
                    if typeName in structs:
                        found = True
                        
                elif argType == "enum":
                    if typeName in enums:
                        found = True
                        
                elif argType == "union":
                    if typeName in unions:
                        found = True
                        
                if not found:
                    raise Exception(f"""Error #B004:  {argType} {typeName} not found in files """)
                else:
                    logging.debug(f"""definition ok: {argType} {typeName} """)

def assertDifferentNamesOcallEcall(ecallFunctions: {}, ocallFunctions: {}):      
        sameFunctionNames = set(ecallFunctions.keys()).intersection(set(ocallFunctions.keys())) # the ecall and ocall with same name
        if sameFunctionNames:
            raise Exception("Error #B005: ocall and ecall must have different names!" + str (sameFunctionNames))
        else:
            logging.debug("check unique names: ok!")


def copyRemoveAndCopyTree(src: str, dst:str = "generated_trusted"):
    userDirName: str = "application"
    userDir = f"{dst}/{userDirName}"

    logging.info(f"Copying all source three in: {userDir}")
    try:
        if os.path.exists(dst) and os.path.isdir(dst):  #todo remove comment
            shutil.rmtree(dst)
        
        shutil.copytree(src, userDir)
        os.chdir(userDir)
        logging.info(f"Working on: {userDir}")


    except Exception as e:
        logging.error(e)
        raise e


def checkFilesExistance(filesPath: []):
    """Check the existance of each file"""
    logging.debug("Check files existance")
    for filePath in filesPath:
        if os.path.exists(filePath) and os.path.isfile(filePath):
            logging.debug(f"""file: {filePath} exists""")
        else:
            raise Exception(f"""Error #B006: file {filePath} not found""")


def checkAllFileCOrH(filePaths: []):
    logging.debug("Check file extension")
    for filePath in filePaths:
        if not filePath.lower().endswith(('.c', '.h')):
            logging.error(f"""Unknown extension for file {filePath}""")

def checkToolsAvailable():
    """return an exception if the tool are not available in console"""
    ctagsInterface.isToolAvailable()
    framaInterface.isToolAvailable()
    gccInterface.isToolAvailable()
        


#DONE
#main(["bcd.c"], "pacchetti_codice_c/compatibili/bcd_compatibile")
#main(["morse.c"], "pacchetti_codice_c/compatibili/morse_compatibile")
#main(["banner.c"], "pacchetti_codice_c/compatibili/banner_compatibile")
#main(["pom.c"], "pacchetti_codice_c/compatibili/pom_compatibile")
#main(["binclock.c"], "pacchetti_codice_c/compatibili/binclock_compatibile") 
#main(["file1.c", "file1.h"], "pacchetti_codice_c/originali/file_c/")

#main(["bubble_sort.c"], "visual_studio/progetti_untrusted/bubble_sort") 
#main(["bubble_sort.c"], "visual_studio/progetti_untrusted/bubble_sort_10x") 
#main(["ecall_cost.c"], "visual_studio/progetti_untrusted/ecall_cost") 
#main(["function_call_cost.c"], "visual_studio/progetti_untrusted/function_call_cost") 

#main(["big.c"], "test/original/big_test_deep_3") 



#files = ["banner.c"]
#rootFolder =  "pacchetti_codice_c/compatibili/banner_compatibile"

#main(files, rootFolder)



