import logging

class Constants(object):
    @staticmethod
    def writeEnclaveManager(filePath: str = "enclave_manager.h"):
            """ Try to create a new file, override the file if already exists """
            with open(filePath, "w") as file: 
                logging.info("Writing file: " + filePath )
                file.write(Constants.__enclave_manager_code__)


    
    __enclave_manager_code__ =  r"""
#include <stdio.h>
#include <stdlib.h>

#include "Enclave_u.h"

#ifdef _MSC_VER
#else
#include <sys/sysinfo.h>
#include <unistd.h>
#endif

#include "sgx_eid.h"
#include "sgx_urts.h"
#include "sgx_utils.h"
#include "sgx_capable.h"


#ifndef _ENCLAVE_MANAGER_
#define _ENCLAVE_MANAGER_


#ifdef _MSC_VER
#define ENCLAVE_NAME "Enclave.signed.dll"
#else
#define ENCLAVE_NAME "enclave.signed.so"
#endif

#define check_hardware_capable 0
  
void create_enclave (void); 
void destroy_enclave (void); 
  
int query_and_enable_sgx_device();

sgx_enclave_id_t my_enclave_id = 0;

/* create the enclave, exit there is problems */
/* if the enclave alread exists, just return and do nothing*/

void create_enclave(void){

	if(my_enclave_id != 0){
		return;
	}

    int update = 0;
    sgx_launch_token_t token = {0};

#if check_hardware_capable
    if(query_and_enable_sgx_device()==-1){
        //compatibility problem
        //default error handling
        printf("Compatibility problem.\n");
        exit(-1);
    } 
#endif
    
    // load enclave
    if (SGX_SUCCESS != sgx_create_enclave(ENCLAVE_NAME, SGX_DEBUG_FLAG, &token, &update, &my_enclave_id, NULL)) {
        
        //default error handling
        printf("failed to load enclave.\n");
        exit(-2);
    }

    // succeed to load enclaves. 
    return;
}

/* return if ok, exit if error */ 
void destroy_enclave(void){
    if(my_enclave_id == 0)
        return;

    if(SGX_SUCCESS != sgx_destroy_enclave(my_enclave_id))
        exit(-1);
	
	my_enclave_id = 0;
    //manage the error if needed
}


#if check_hardware_capable
/* name: query_and_enable_sgx_device()
 * Usage: 1. It detects SGX device's status inside target environment.
 *        2. And it will enable SGX device when it detects that SGX was set as Software Control inside BIOS. 
 *           So it can be used inside desktop application installer to enable SGX in target environment during installation.
 * Note: This function need be executed with Administrator privilege.
*/
int query_and_enable_sgx_device()
{
	sgx_device_status_t sgx_device_status;
	sgx_status_t sgx_ret = sgx_cap_get_status(&sgx_device_status);
	if (sgx_ret != SGX_SUCCESS) {
		printf("Failed to get SGX device status. If you have problem, run with administrator privilege\n");
		return -2;
	}
	else {
		switch (sgx_device_status) {
		case SGX_ENABLED:
			printf("SGX device has been enabled.\n");
            return 0; //ok
		
        case SGX_DISABLED_REBOOT_REQUIRED:
			printf("SGX device has been enabled. Please reboot your machine.\n");
			break;
		case SGX_DISABLED_LEGACY_OS:
			printf("SGX device can't be enabled on an OS that doesn't support EFI interface.\n");
			break;
		case SGX_DISABLED:
			printf("SGX is not enabled on this platform. More details are unavailable.\n");
			break;
		case SGX_DISABLED_MANUAL_ENABLE:
			printf("SGX device can be enabled manually in the BIOS setup.\n");
			break;
		case SGX_DISABLED_HYPERV_ENABLED:
			printf("Detected an unsupported version of Windows* 10 with Hyper-V enabled.\n");
			break;
		case SGX_DISABLED_UNSUPPORTED_CPU:
			printf("SGX is not supported by this CPU.\n");
			break;
		default:
			printf("Unexpected error.\n");
			break;

		case SGX_DISABLED_SCI_AVAILABLE:
			printf("SGX device can be enabled by a Software Control Interface, so enable it\n");
			sgx_ret = sgx_cap_enable_device(&sgx_device_status);
			if (sgx_ret != SGX_SUCCESS) {
				printf("Failed to Enable SGX device.\n");
			}
			switch (sgx_device_status)
			{
			case SGX_ENABLED:
				printf("SGX device has been enabled by Software Control Interface.\n");
				return 0; //ok

			case SGX_DISABLED_REBOOT_REQUIRED:
				printf("SGX device has been enabled by Software Control Interface. Please reboot your machine.\n");
				break;
			case SGX_DISABLED_LEGACY_OS:
				printf("SGX device can't be enabled on an OS that doesn't support EFI interface.\n");
				break;
			case SGX_DISABLED:
				printf("SGX is not enabled on this platform. More details are unavailable.\n");
				break;
			case SGX_DISABLED_MANUAL_ENABLE:
				printf("SGX device can be enabled manually in the BIOS setup.\n");
				break;
			case SGX_DISABLED_HYPERV_ENABLED:
				printf("Detected an unsupported version of Windows* 10 with Hyper-V enabled.\n");
				break;
			case SGX_DISABLED_UNSUPPORTED_CPU:
				printf("SGX is not supported by this CPU.\n");
				break;
			default:
				printf("Unexpected error.\n");
                break;
			}
			break;
		}
	}
	return -1;
}
#endif /* check_hardware_capable */


#endif /* !_ENCLAVE_MANAGER_ */




  
"""